/*
BibliMath - Bibliothèque mathématique

Copyright (C) 2012 Observatoire de Paris-Meudon

Ce programme est un logiciel libre ; vous pouvez le redistribuer et/ou le modifier conformément aux dispositions de la Licence Publique Générale GNU, telle que publiée par la Free Software Foundation ; version 2 de la licence, ou encore (à votre choix) toute version ultérieure.

Ce programme est distribué dans l'espoir qu'il sera utile, mais SANS AUCUNE GARANTIE ; sans même la garantie implicite de COMMERCIALISATION ou D'ADAPTATION A UN OBJET PARTICULIER. Pour plus de détail, voir la Licence Publique Générale GNU .

Vous devez avoir reçu un exemplaire de la Licence Publique Générale GNU en même temps que ce programme ; si ce n'est pas le cas, écrivez à la Free Software Foundation Inc., 675 Mass Ave, Cambridge, MA 02139, Etats-Unis.
*/


/** Bibliothèque mathématique : fonctions remarquables etc.
*/

BibliMath = function() {

return {


/** Calcule le logarithme népérien de la fonction gamma avec x comme argument.<br>
    Si x <= 6, alors on utilise la relation de récurrence pour rendre x inférieur à 3
    puis on applique l'approximation rationnelle 5236 de Hart et al, 
    Computer Approximations, John Wiley and Sons, NY, 1968.<br>
    Si x > 6, on utilise la relation de récurrence pour rendre x supérieur à 12
    puis on applique l'approximation rationnelle 5423 de la même source.
*/
    //public static double alngam(double x) {
    alngam: function(x) {
        var hln2pi, offset, prod, xx, resu, n, scoefn, scoefd, coef, i;
        
        hln2pi = 0.91893853320467274178;
        
        scoefn = [ 0.62003838007127258804E2,
                    0.36036772530024836321E2,
                    0.20782472531792126786E2,
                    0.6338067999387272343E1,
                    0.215994312846059073E1,
                    0.3980671310203570498E0,
                    0.1093115956710439502E0,
                    0.92381945590275995E-2,
                    0.29737866448101651E-2 ];
        
        scoefd = [ 0.62003838007126989331E2,
                    0.9822521104713994894E1,
                    -0.8906016659497461257E1,
                    0.1000000000000000000E1 ];
        
        coef = [ 0.83333333333333023564E-1,
                  -0.27777777768818808E-2,
                  0.79365006754279E-3,
                  -0.594997310889E-3,
                  0.8065880899E-3 ];
        
        // x est plus petit ou égal à 6
        if ( x <= 6.0 ) {
            prod = 1.;
            xx = x;
            if ( x > 3.0 ) {
                while ( xx > 3.0 ) {
                    xx = xx - 1.;
                    prod = prod*xx;
                }
            }
            if (x < 2.0) {
                while ( xx < 2.0 ) {
                    prod = prod*xx;
                    xx = xx + 1.;
                }
            }
            
            // on calcule gamma(x) par une approximation rationnelle
            resu = this.devlpl(scoefn, 9, xx - 2.0) / this.devlpl(scoefd, 4, xx - 2.0);
            
            resu = resu * prod;
            resu = Math.log(resu);
            return resu;

        } else { // x est plus grand que 6
            
            offset = hln2pi;
            
            n = Math.floor(12.0 - x);
            if ( n <= 0 ) {
                xx = x;
            } else {
                prod = 1.;
                for ( i = 1 ; i <= n ; i++ ) {
                    prod = prod * (x + (i - 1));
                }
                offset = offset - Math.log(prod);
                xx = x + n;
            }

            // on calcule le développement en série
            resu = this.devlpl(coef, 5, 1. / Math.pow(xx, 2.)) / xx;
            resu = resu + offset + (xx - 0.5) * Math.log(xx) - xx;
            return resu;
        }
    },

    /** Calcule la valeur d'un polynôme de degré n-1 au point x. Les coefficients 
    sont donnés dans le tableau a, soit<br>
    a[0] + a[1]*x + ... + a[n-1]*x^(n-1)
    */
    //public static double devlpl(double a[], int n, double x) {
    devlpl: function(a, n, x) {
        var term, i;
        
        term = a[n-1];
        for ( i = n - 2 ; i >= 0 ; i-- ) {
            term = a[i] + term * x;
        }
        
        return term;

    },

    /** Transformée de Fourier 2D. Calcule une TF 1D selon l'une des directions
    puis une autre sur le tableau obtenu selon l'autre direction.
    */
    //public static void FFTcmplx2cmplx2D(int sign, int n, double ar[][], double ai[][]) {
    FFTcmplx2cmplx2D: function(sign, n, ar, ai) {
        var tempoRe, tempoIm, i, j;
        
        tempoRe = new Array(n);
        tempoIm = new Array(n);

        // TF 1D selon la première direction (colonnes en l'occurrence)
        for ( i = 0 ; i < n ; i++ ) {
            for ( j = 0 ; j < n ; j++ ) {
                tempoRe[j] = ar[i][j];
                tempoIm[j] = ai[i][j];
            }
            this.FFTcmplx2cmplx(sign, n, tempoRe, tempoIm);
            for ( j = 0 ; j < n ; j++ ) {
                ar[i][j] = tempoRe[j];
                ai[i][j] = tempoIm[j];
            }
        }

        // TF 1D selon la deuxième direction (lignes)
        for ( j = 0 ; j < n ; j++ ) {
            for ( i = 0 ; i < n ; i++ ) {
                tempoRe[i] = ar[i][j];
                tempoIm[i] = ai[i][j];
            }
            this.FFTcmplx2cmplx(sign, n, tempoRe, tempoIm);
            for ( i = 0 ; i < n ; i++ ) {
                ar[i][j] = tempoRe[i];
                ai[i][j] = tempoIm[i];
            }
        }
    },

    /** Transformée de Fourier 1D.
    */
    //public static void FFTcmplx2cmplx(int sign, int n, double ar[], double ai[]) {
    FFTcmplx2cmplx: function(sign, n, ar, ai) {
        var scale, i, j, tempr, tempi, m, mmax, istep, delta, w, wr, wi, tr, ti;
        
        scale = Math.sqrt(1.0 / n);
        
        for (i=j=0; i<n; ++i) {
            if (j >= i) {
                tempr = ar[j] * scale;
                tempi = ai[j] * scale;
                ar[j] = ar[i] * scale;
                ai[j] = ai[i] * scale;
                ar[i] = tempr;
                ai[i] = tempi;
            }
            m = n / 2;
            while (m>=1 && j>=m) {
                j -= m;
                m /= 2;
            }
            j += m;
        }
        
        for (mmax=1, istep=2*mmax; mmax<n; mmax=istep, istep=2*mmax) {
            delta = sign * Math.PI / mmax;
            for (m=0; m<mmax; ++m) {
                w = m * delta;
                wr = Math.cos(w);
                wi = Math.sin(w);
                for (i=m; i<n; i+=istep) {
                    j = i + mmax;
                    tr = wr * ar[j] - wi * ai[j];
                    ti = wr * ai[j] + wi * ar[j];
                    ar[j] = ar[i] - tr;
                    ai[j] = ai[i] - ti;
                    ar[i] += tr;
                    ai[i] += ti;
                }
            }
            mmax = istep;
        }
    }

};

}();
